"""Interactive grading tool"""
from __future__ import annotations
import os
import shutil
import curses
import traceback
import argparse

from copy import deepcopy

from .commands import parse_grading_schema, collect_test_scripts, TMP_TESTS_DIR
from .model import GradingSchema
from ..test_runner import run_tests_for, load_tests, GraderTestResult

QUIT_KEY = 'q'
ESCAPE_KEY = 27

GREEN_COLOR = None


def run_tests_interactive_argument_parser(argument_parser: argparse.ArgumentParser) -> None:
    """Add necessary arguments for interactive grading"""
    argument_parser.add_argument('-gs', '--grading-scheme', required=False)


def get_students_folders() -> list[str]:
    """Return a list of student folders from current directory"""
    return [path for path in os.listdir() if os.path.isdir(path) and
            os.path.exists(os.path.join(path, '.submission.yaml'))]


class Panel:

    def __init__(self, window, storage: Model):
        self.window = window
        self.storage = storage
        self.focused = False
        self.rows, self.cols = self.window.getmaxyx()


def is_enter(c: int) -> bool:
    """Check if enter was entered"""
    return c == curses.KEY_ENTER or c == 10 or c == 13


class StudentSidebar(Panel):

    NAME = "student_sidebar"

    def __init__(self, screen, storage):
        super().__init__(screen, storage)

        self.position = 0

        self.width = len(max(self.storage.students_folders, key=len)) + 2
        self.width = max(self.width, len(max(self.storage.problems, key=len)) + 15)

    def draw(self):

        self.window.addstr(0, 1, "STUDENTS:", curses.A_UNDERLINE)

        self.window.vline(0, self.width, curses.ACS_VLINE, 1)
        self.window.vline(1, self.width, curses.ACS_SSSB, 1)
        self.window.vline(2, self.width, curses.ACS_VLINE, self.rows - SummaryPanel.HEIGHT - 2)

        for i, student in enumerate(self.storage.students_folders):
            if i == self.position and self.focused:
                mode = curses.A_REVERSE
            elif i == self.position and not self.focused:
                mode = GREEN_COLOR | curses.A_ITALIC
            else:
                mode = curses.A_NORMAL
            self.window.addstr(i + 1, 1, student, mode)

    def control(self, c):
        if c == curses.KEY_UP and self.position > 0:
            self.position -= 1
        elif c == curses.KEY_DOWN and self.position < len(self.storage.students_folders) - 1:
            self.position += 1
        elif is_enter(c):
            self.storage.set_current_panel(ProblemsTabs.NAME)

        self.storage.current_student = self.storage.students_folders[self.position]


class ProblemsTabs(Panel):
    NAME = 'problem_tabs'

    def __init__(self, window, storage):
        super().__init__(window, storage)

        self.position = 0

        self.problems = [rule.filename for rule in storage.grading_schema.rules]

    def draw(self):

        padding = self.storage.panels[StudentSidebar.NAME].width

        self.window.addstr(0, padding + 2, "FILES:", curses.A_UNDERLINE)
        self.window.hline(1, padding, curses.ACS_HLINE, self.cols - 3)

        padding += 7

        for i, student in enumerate(self.problems):
            if i == self.position and self.focused:
                mode = curses.A_REVERSE
            elif i == self.position and not self.focused:
                mode = GREEN_COLOR | curses.A_ITALIC
            else:
                mode = curses.A_NORMAL

            self.window.addstr(0, padding + 2, student, mode)
            padding += len(student) + 1

    def control(self, c):
        match c:
            case curses.KEY_LEFT if self.position > 0:
                self.position -= 1
            case curses.KEY_RIGHT if self.position < len(self.problems) - 1:
                self.position += 1
            case curses.KEY_ENTER | 10 | 13:
                self.storage.run_test_for_selected_problem()
                self.storage.set_current_panel(TestsResultPanel.NAME)
                self.storage.panels[TestsResultPanel.NAME].reset_cursor()

        self.storage.current_problem_file = self.problems[self.position]


class TestsResultPanel(Panel):
    NAME = 'tests_result_panel'
    TOP_MARGIN = 2

    def __init__(self, window, storage):
        super().__init__(window, storage)

        self.start_y = 0
        self.start_x = 0
        self.margin = 0

        self.visible_vertical_range = self.rows - 3
        self.visible_horizontal_range = 0

        self.x, self.y = 0, 0

        self.lines = None

    def draw(self):

        self.margin = self.storage.panels[StudentSidebar.NAME].width + 2
        self.visible_horizontal_range = self.cols - self.margin

        self.lines = self.storage.get_current_test_output().split('\n')

        for i, line in enumerate(self.lines[self.start_y: self.start_y + self.visible_vertical_range]):
            self.window.addstr(2 + i, self.margin, line[self.start_x:self.start_x + self.visible_horizontal_range])

        if self.focused:
            curses.curs_set(1)
            self.window.move(self.TOP_MARGIN + self.y, self.margin + self.x)
        else:
            curses.curs_set(0)

    def control(self, c):

        length_of_current_line = len(self.lines[self.start_y + self.y])
        max_horizontal_scroll = max(0, length_of_current_line - self.visible_horizontal_range)
        maximum_x = min(self.visible_horizontal_range - 1, length_of_current_line)
        maximum_y = min(len(self.lines), self.visible_vertical_range)

        match c:
            case curses.KEY_UP if self.y > 0:
                self.y -= 1
            case curses.KEY_DOWN if self.y < maximum_y - 1:
                self.y += 1
            case curses.KEY_RIGHT if self.x < maximum_x:
                self.x += 1
            case curses.KEY_LEFT if self.x > 0:
                self.x -= 1
            case curses.KEY_DOWN if self.y == maximum_y - 1:
                self.start_y = min(self.start_y + 1, len(self.lines) - maximum_y - 1)
            case curses.KEY_UP if self.y == 0:
                self.start_y = max(self.start_y - 1, 0)
            case curses.KEY_RIGHT if self.x == maximum_x:
                self.start_x = min(self.start_x + 1, max_horizontal_scroll)
            case curses.KEY_LEFT if self.x == 0:
                self.start_x = max(self.start_x - 1, 0)
            case 36:  # $
                self.start_x = max(0, length_of_current_line - self.visible_horizontal_range)
                self.x = maximum_x
            case 94:  # ^
                self.x = 0
                self.start_x = 0
            case 71:  # G
                self.start_y = max(0, len(self.lines) - maximum_y)
                self.y = maximum_y - 1
            case 103:  # g
                self.start_y = 0
                self.y = 0
            case 114:  # r
                self.storage.run_test_for_selected_problem()

    def reset_cursor(self):
        self.x, self.y = 0, 0


class SummaryPanel(Panel):

    NAME = 'summary_panel'
    HEIGHT = 15

    def __init__(self, window, storage):
        super().__init__(window, storage)

    def draw(self):

        width = self.storage.panels[StudentSidebar.NAME].width
        start_y = self.rows - self.HEIGHT

        self.window.hline(start_y, 0, curses.ACS_HLINE, width)
        self.window.addch(start_y, width, curses.ACS_SBSS)
        self.window.vline(start_y + 1, width, curses.ACS_VLINE, self.HEIGHT)

        self.window.addstr(start_y + 1, 1, "SUMMARY:", curses.A_UNDERLINE)

        total_grade = 0

        for i, rule in enumerate(self.storage.grading_schema.rules):
            test_result = self.storage.test_runner_result[self.storage.current_student].get(rule.filename)

            if test_result is not None:
                max_grade = self.storage.grading_schema.total_grade * rule.weight / 100
                gained_grade = round(test_result.gained_points / test_result.total_points * max_grade, 2)
            else:
                max_grade, gained_grade = 0, 0

            total_grade += gained_grade

            result_string = "No Data" if test_result is None else f"{gained_grade}/{max_grade}"

            self.window.addstr(start_y + i + 2, 1, f" - {rule.filename}: {result_string}")

        self.window.addstr(start_y + len(self.storage.grading_schema.rules) + 2, 2,
                           f"> Total Grade: {total_grade}/{self.storage.grading_schema.total_grade}")

class Model:
    """Class for storing global data of app"""

    def __init__(self, grading_schema: GradingSchema):
        self.current_panel = None
        self.panels = {}

        self.grading_schema = grading_schema

        self.students_folders = get_students_folders()
        self.problems = [rule.filename for rule in self.grading_schema.rules]

        self.current_student = self.students_folders[0]
        self.current_problem_file = grading_schema.rules[0].filename

        self.test_runner_result = {student: {} for student in self.students_folders}
        self.test_cases = {}

        self.load_test_cases()

    def load_test_cases(self) -> None:
        """Load test cases from specified locations in grading schema and write them to self.test_cases"""
        collected_test_scripts = collect_test_scripts(self.grading_schema)

        for rule, copied_path in collected_test_scripts.items():
            self.test_cases[rule] = load_tests(copied_path)

        shutil.rmtree(TMP_TESTS_DIR)

    def add_panel(self, name: str, panel) -> None:
        """Add panel to panel registry"""
        self.panels[name] = panel

    def set_current_panel(self, name: str) -> None:
        """Set current panel"""
        if self.current_panel is not None:
            self.current_panel.focused = False

        self.current_panel = self.panels.get(name)
        self.current_panel.focused = True

    def get_current_test_output(self) -> str:
        if self.current_problem_file not in self.test_runner_result[self.current_student]:
            return ""

        return self.test_runner_result[self.current_student][self.current_problem_file].output

    def get_current_test_result(self) -> GraderTestResult | None:
        if self.current_problem_file not in self.test_runner_result[self.current_student]:
            return None

        return self.test_runner_result[self.current_student][self.current_problem_file]

    def run_test_for_selected_problem(self) -> None:
        script_path = os.path.join(self.current_student, self.current_problem_file)

        grading_rule = next(filter(lambda r: r.filename == self.current_problem_file, self.grading_schema.rules))

        result = run_tests_for(script_path=script_path, suite=deepcopy(self.test_cases[grading_rule]))

        self.test_runner_result[self.current_student][self.current_problem_file] = result


def create_color_pairs() -> None:
    """Init color pairs"""
    global GREEN_COLOR
    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)

    GREEN_COLOR = curses.color_pair(1)


def wrapper(screen, args: argparse.Namespace) -> None:

    create_color_pairs()
    storage = Model(parse_grading_schema(args.grading_schema))

    sidebar = StudentSidebar(screen, storage)
    problem_tabs = ProblemsTabs(screen, storage)
    summary_panel = SummaryPanel(screen, storage)
    test_result_panel = TestsResultPanel(screen, storage)

    # Must be in such order to maintain proper drawing
    storage.add_panel(ProblemsTabs.NAME, problem_tabs)
    storage.add_panel(SummaryPanel.NAME, summary_panel)
    storage.add_panel(StudentSidebar.NAME, sidebar)
    storage.add_panel(TestsResultPanel.NAME, test_result_panel)

    storage.set_current_panel(StudentSidebar.NAME)

    sidebar.draw()
    problem_tabs.draw()
    summary_panel.draw()

    while True:
        screen.erase()

        for panel in storage.panels.values():
            panel.draw()

        c = screen.getch()

        match c:
            case 113:
                break
            case 97:
                storage.set_current_panel(StudentSidebar.NAME)
            case 115:
                storage.set_current_panel(ProblemsTabs.NAME)
            case 100:
                storage.set_current_panel(TestsResultPanel.NAME)

        storage.current_panel.control(c)


def run_tests_interactive(args: argparse.Namespace) -> None:
    """Run the test with terminal interface"""
    try:
        curses.wrapper(wrapper, args)
    except Exception as ex:
        traceback.print_exc()


"""Util functions"""
import os
import sys
import zipfile
from pathlib import Path
from platform import system
from subprocess import Popen, PIPE, run

import yaml

from .classroom.model import Course, CourseWork, Submission, Student, Attachment


def capture_io(script_name: str, inputs: list) -> str:
    """
    :param script_name:  name of the script to be tested
    :param inputs: list of inputs, one element per line
    :return: returns script output for the test case
    """

    inputs = "\n".join(inputs)

    environ = os.environ.copy()
    environ['PYTHONIOENCODING'] = 'utf-8'

    p = Popen(
        [sys.executable, script_name],
        stdin=PIPE,
        stdout=PIPE,
        env=environ,
        stderr=PIPE,
        encoding="utf8")

    stdout, stderr = p.communicate(input=inputs)

    return stdout.strip()


def extract_zip(source: str, destination: str) -> None:
    """
    Extract zip file at source to destination
    :param source:
    :param destination:
    :return:
    """
    with zipfile.ZipFile(source, 'r') as output:
        output.extractall(destination)


def create_unique_dir(parent_dir: str, new_dir_name: str) -> str:
    """
    creates a directory named new_dir_name
    if such a directory exists, it new directory will be created with unique name
    """
    os.chdir(parent_dir)

    if os.path.exists(new_dir_name) and os.listdir(new_dir_name):
        i = 1
        while os.path.exists(f"{new_dir_name}_{i}") and os.listdir(f"{new_dir_name}_{i}"):
            i += 1

        new_path = f"{new_dir_name}_{i}"

        os.mkdir(new_path)

        return new_path

    os.mkdir(new_dir_name)

    return new_dir_name


def get_current_course() -> Course:
    """
    Searches for .course.yaml file in current directory and returns the course entity from it
    :return:
    :raises: FileNotFoundError if .course.yaml file is not found
    """
    current_dir = [file for file in os.listdir('.') if os.path.isfile(file) and file == ".course.yaml"]
    if '.course.yaml' not in current_dir:
        raise FileNotFoundError("Can't find .course.yaml in this directory")

    with open('.course.yaml', 'r', encoding='utf8') as course_file:
        course = yaml.safe_load(course_file.read())

    return Course(id=course['id'], name=course['name'])


def find_descendant(file_name: str) -> str:
    """
    looks in previous paths for a file. returns a path to a file
    """

    path = Path(os.getcwd())

    while not os.path.exists(os.path.join(path, file_name)) and not os.path.ismount(path):
        path = path.parent.absolute()

    return_path = os.path.join(path, file_name)

    if not os.path.exists(return_path):
        raise FileNotFoundError(file_name)

    return os.path.join(path, file_name)


def write_course_yaml(course: Course, path: str) -> None:
    with open(os.path.join(path, ".course.yaml"), "w", encoding="utf-8") as f:
        yaml.dump(course.__dict__, f)


def read_course_yaml(path: str) -> Course:
    with open(os.path.join(path, ".course.yaml"), "r", encoding="utf-8") as f:
        course_info = yaml.safe_load(f)

    return Course(id=course_info["id"], name=course_info["name"], description=course_info["description"],
                  section=course_info["section"])


def write_course_work_yaml(course_work: CourseWork, path: str) -> None:
    with open(os.path.join(path, ".course_work.yaml"), "w", encoding="utf-8") as f:
        yaml.dump(course_work.__dict__, f)


def read_course_work_yaml(path: str) -> CourseWork:
    with open(os.path.join(path, ".course_work.yaml"), "r", encoding="utf-8") as f:
        course_work_info = yaml.safe_load(f)

    return CourseWork(id=course_work_info["id"],
                      title=course_work_info['title'],
                      description=course_work_info['description'],
                      course_id=course_work_info['course_id'],
                      link=course_work_info['link'])


def read_sumbission_yaml(path: str) -> Submission:
    with open(os.path.join(path, ".submission.yaml"), "r", encoding="utf-8") as f:
        submission_info = yaml.safe_load(f)

    return Submission(id=submission_info["id"], user_id=submission_info["user_id"],
                      course_id=submission_info["course_id"], coursework_id=submission_info["coursework_id"],
                      creation_time=submission_info["creation_time"], update_time=submission_info["update_time"],
                      attachments=[Attachment(id=value["id"], title=key, link=value["link"]) for key, value in
                                   submission_info["attachments"].items()], link=submission_info["link"])


def write_submission_yaml(submission: Submission, student: Student, path: str) -> None:
    with open(os.path.join(path, ".submission.yaml"), "w", encoding="utf-8") as f:
        submission_dict = submission.__dict__.copy()
        submission_dict["attachments"] = {attachment.title: {"id": attachment.id, "link": attachment.link} for
                                          attachment in submission_dict["attachments"]}
        submission_dict["student_name"] = student.name
        yaml.dump(submission_dict, f, allow_unicode=True)






import argparse
import inspect
import os
import sys
import unittest
from io import BufferedIOBase
from typing import Type
from unittest import TextTestRunner, TestSuite, TestLoader

from six import StringIO

from .converter import create_unittest
from .globals import set_testing_script_path, get_testing_script_path


class GraderTestResult(unittest.TextTestResult):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.total_points = None
        self.gained_points = None

        self.output = None


class GraderTestRunner(TextTestRunner):
    """Test runner that counts points for each successful test case."""

    def run(self, test_suite: TestSuite) -> GraderTestResult:
        """Run tests and count total grade"""

        test_cases = self._discover_test_cases(test_suite)

        self.stream.writeln(f"=== Running tests for: {get_testing_script_path()} ===")

        result = super().run(test_suite)

        self.show_points(test_cases, result)

        self.stream.seek(0)
        result.output = self.stream.read()

        return result

    def show_points(self, test_cases: list[Type], result: GraderTestResult) -> None:
        """
        Count and print gained points
        Args:
            test_case: class of testcase, we must pass it because, after running tests
            it is very hard to find TestCase class
            result: Test result of the TestCase
        """

        all_methods = [method
                       for test_case in test_cases
                       for name, method in inspect.getmembers(test_case)
                       if name.startswith("test_")]

        failed_methods_names = [method._testMethodName for method, _ in result.failures] + \
                               [method._testMethodName for method, _ in result.errors]
        succeeded_methods = [method for method in all_methods if method.__name__ not in failed_methods_names]

        total_points = sum(getattr(method, "points_increment") for method in all_methods)
        gained_points = sum(getattr(method, "points_increment") for method in succeeded_methods)

        self.stream.writeln(f"Grade: {gained_points}/{total_points}")

        result.total_points = total_points
        result.gained_points = gained_points

    @staticmethod
    def _discover_test_cases(test_suite: TestSuite) -> set:
        """Get all TestCase classes from a TestSuite
        Args:

            test_suite: suite where to search for TestCase classes
        """
        discovered_test_cases = set()

        nodes_to_discover = [test_suite]
        while nodes_to_discover:
            node = nodes_to_discover.pop()

            for test in node._tests:
                if isinstance(test, TestSuite):
                    nodes_to_discover.append(test)
                else:
                    discovered_test_cases.add(test.__class__)

        return discovered_test_cases


def load_python_tests(test_file_path: str) -> TestSuite:
    """Load test suite from file by path

    Args:
        test_file_path: path to the test script

    """
    directory_path = os.path.dirname(test_file_path)
    test_file_name = os.path.basename(test_file_path)

    loader = TestLoader().discover(start_dir=directory_path, top_level_dir=directory_path,
                                   pattern=test_file_name)
    return loader


def load_tests(test_script_path: str) -> TestSuite:
    """Load test suite from python or .tests file"""

    if test_script_path.endswith(".py"):
        suite = load_python_tests(test_script_path)
    elif test_script_path.endswith(".tests"):
        suite = TestLoader().loadTestsFromTestCase(create_unittest(test_script_path))
    else:
        raise TypeError("Invalid test file extension")

    return suite


def run_tests_setup_argparse(argument_parser: argparse.ArgumentParser) -> None:
    """Add necessary arguments to argparse"""
    argument_parser.add_argument("-t", "--tests", help="Path to the test script")
    argument_parser.add_argument("-s", "--solution", help="Path to the solution")


def run_tests_for(script_path: str, test_script_path: str = None, suite: TestSuite = None) -> GraderTestResult:
    """
    Run tests for given script with test_script_path
    :param script_path:
    :param test_script_path:
    :param suite:
    :return:
    """

    if test_script_path is None and suite is None:
        raise ValueError("You must provide either test_script_path or suite")

    set_testing_script_path(script_path)

    if suite is None:
        suite = load_tests(test_script_path)

    stream = StringIO()
    runner = GraderTestRunner(stream)

    result = runner.run(suite)

    return result


def run_tests(args: argparse.Namespace) -> None:
    """Perform run tests action"""

    test_file_path = args.tests
    solution_file_path = args.solution

    result = run_tests_for(script_path=solution_file_path, test_script_path=test_file_path)

    result.stream.seek(0)
    print(result.stream.read())



"""Singletone Google Services"""
import os.path

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow

from .classroom.connector import ClassroomConnector
from .drive import GoogleDriveService
from .utils import find_descendant

SCOPES = ClassroomConnector.SCOPES + GoogleDriveService.SCOPES


class Auth:

    def __init__(self, credentials_file: str):
        """"""
        self.credentials_file = credentials_file

    def get_credentials(self):
        """Return credentials for all Google Services"""
        creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        try:

            token_path = find_descendant("token.json")

            creds = Credentials.from_authorized_user_file(token_path, SCOPES)
            # If there are no (valid) credentials available, let the user log in.
        except FileNotFoundError:
            print("token.json not found")
        finally:
            if not creds or not creds.valid:
                if creds and creds.expired and creds.refresh_token:
                    creds.refresh(Request())
                else:
                    flow = InstalledAppFlow.from_client_secrets_file(
                        self.credentials_file, SCOPES
                    )
                    creds = flow.run_local_server(port=0)
                # Save the credentials for the next run
                with open("token.json", "w") as token:
                    token.write(creds.to_json())

        return creds


auth = Auth(find_descendant("credentials.json"))
credentials = auth.get_credentials()

drive_service = GoogleDriveService(credentials=credentials)
classroom_service = ClassroomConnector(credentials=credentials)



import argparse
import os
import shutil
from argparse import ArgumentParser

import inquirer
import yaml

from .model import Course, CourseWork, GradingRule, GradingResult, Student
from ..classroom.model import Submission, GradingSchema
from ..services import drive_service, classroom_service
from ..test_runner import run_tests_for
from ..utils import create_unique_dir, extract_zip, get_current_course, write_course_yaml, read_course_yaml, \
    write_course_work_yaml, read_sumbission_yaml, write_submission_yaml, read_course_work_yaml


def run_no_args_argparse(argument_parser: argparse.ArgumentParser) -> None:
    """Add necessary arguments to argparse"""
    # argument_parser.add_argument()
    pass


def get_course(*args):
    courses = classroom_service.get_courses()

    course_choices = {course.name: course for course in courses}

    questions = [
        inquirer.List('course', message="Select course", choices=list(course_choices.keys())),
    ]

    selected_course = course_choices[inquirer.prompt(questions)['course']]

    new_path = create_unique_dir(os.getcwd(), selected_course.name)

    write_course_yaml(selected_course, new_path)


def choose_course_work_argparse(argument_parser: argparse.ArgumentParser) -> None:
    """Setup necessary arguments for downloading course"""
    argument_parser.add_argument('-stf', '--students-file', required=False,
                                 help="Path to the file where names of students whose assignments will be downloaded")
    argument_parser.add_argument("-eisd", "--extract_if_single_dir", required=False, help="Extract if single directory",
                                    action="store_true")


def get_students_from_file(course: Course, student_file_path: str) -> dict[str, Student]:
    """
    Read students names from file by path and return their Student objects

    :param course: course object
    :param student_file_path: student file path
    :return: dict of student objects from that file
    """

    with open(student_file_path, "r", encoding="utf8") as student_file:
        students_names = list(map(str.strip, student_file.readlines()))

    students_list = list(filter(lambda s: s.name in students_names,
                                classroom_service.get_students(course_id=course.id).values()))

    return {student.id: student for student in students_list}


def choose_course_work(args: argparse.Namespace) -> None:
    """
    Download course work from classroom using terminal select

    :param args: args from command line
    :return: None
    """

    if not os.path.exists(".course.yaml"):
        raise FileNotFoundError(".course.yaml not found. What course do you want?")

    course_info = read_course_yaml(os.getcwd())

    course_works = {assignment.title: assignment for assignment in classroom_service.get_courseworks(course_info.id)}

    questions = [
        inquirer.List('assignment', message="Select assignment", choices=list(course_works.keys())),
    ]

    selected_course = course_works[inquirer.prompt(questions)['assignment']]

    if args.students_file is not None:
        students = get_students_from_file(course=course_info, student_file_path=args.students_file)
    else:
        students = classroom_service.get_students(course_id=course_info.id)

    get_course_work(course_info, selected_course, students=students, extract_if_single_directory=args.extract_if_single_dir)


def get_submission(
    submission: Submission,
    target_dir: str,
    student: Student,
    extract_if_single_directory: bool = True,
) -> None:
    """
    Download submission attachments from specified submission directory.
    All attachments are stored in directory with target_dir (if it doesn't exist, it will be created)
    Also creates additional file .submission.yaml inside that folder to store submission objects
    If there already exists .submission.yaml in the target_dir, it will update the submission attachments, if
    they has been updated

    :param submission: submission to download
    :param target_dir: where to download the submission
    :param student: student of submission
    :param extract_if_single_directory: if True, extract zip file if it contains only one directory
    :return: None
    """
    attachments_updated = True

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    if os.path.exists(os.path.join(target_dir, ".submission.yaml")):
        submission_info = read_sumbission_yaml(target_dir)

        if submission_info.update_time == submission.update_time:
            attachments_updated = False
        else:
            write_submission_yaml(submission, student, target_dir)
    else:
        write_submission_yaml(submission, student, target_dir)

    if attachments_updated:
        for attachment in submission.attachments:
            target_download_path = os.path.join(target_dir, attachment.title)
            drive_service.download_file(attachment.id, target_download_path)
            if attachment.title.endswith(".zip"):
                extract_zip(target_download_path, target_dir)
                os.remove(target_download_path)

                # if the only thing in the archive was a directory,
                # move all the contents from the directory and remove it
                if not extract_if_single_directory:
                    continue
                contents = os.listdir(target_dir)
                contents = [file for file in contents if not file.startswith(".")]
                if len(contents) != 1:
                    continue
                dir_to_extract = os.path.join(target_dir, contents[0])
                if not os.path.isdir(dir_to_extract):
                    continue
                print(f"Extracting from {dir_to_extract} to {target_dir}")
                for file in os.listdir(dir_to_extract):
                    shutil.move(os.path.join(dir_to_extract, file), target_dir)
                os.rmdir(dir_to_extract)


def get_course_work(
    course_info: Course,
    course_work: CourseWork,
    students: dict = None,
    target_dir: str = ".",
    extract_if_single_directory: bool = True,
):
    """
    Download attachments of submission from specified course_work of specified students to
    target_dir

    :param course_info: course from which to download submissions
    :param course_work: course from which to download submissions
    :param students: students from which to download submissions (if None, all students will be downloaded)
    :param target_dir: target directory to download attachments
    :param extract_if_single_directory: if True, extract zip file if it contains only one directory
    :return:
    """

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    course_work_dir = os.path.join(target_dir, course_work.title)
    if not os.path.exists(course_work_dir):
        os.makedirs(course_work_dir)

    if not os.path.exists(os.path.join(course_work_dir, ".course_work.yaml")):
        write_course_work_yaml(course_work, course_work_dir)

    if students is None:
        students = classroom_service.get_students(course_info.id)

    for submission in classroom_service.get_submissions(course_info.id, course_work.id):
        if submission.user_id not in students:
            continue

        get_submission(submission, os.path.join(course_work_dir, students[submission.user_id].name),
                       students[submission.user_id], extract_if_single_directory)


def setup_grade_submissions(argument_parser: ArgumentParser) -> None:
    """Setup argument parser for using action grade_submissions"""
    argument_parser.add_argument('-cw', '--coursework', help="Id of the coursework from URL in classroom")
    argument_parser.add_argument('-sid', '--submission-id', help="Id of the submission from URL in classroom")
    argument_parser.add_argument('-sf', '--submission-folder', help="Path to the folder with submissions of coursework")
    argument_parser.add_argument('-cg', '--classroom-grade',
                                 help="If this option is provided, automatically grade submission in classroom."
                                      "Otherwise, create csv with grades.", action="store_true")
    argument_parser.add_argument('-e', '--export-file', help="Path to the CSV file with grades.")
    argument_parser.add_argument("-gs", "--grading-schema", help="Path to the grading_schema YAML file")


def get_submissions_to_grade(course: Course, coursework: CourseWork, args: argparse.Namespace) -> list[Submission]:
    """
    Return the list of all submissions that must be graded
    If only course and coursework are provided (no additional data in args), then return all submissions
    If --submission-id is provided, grade only submission with that ID [from URL in classroom web]
    If --submission-folder provided, return all submissions from that folder

    :param course: Course, where to search for submissions
    :param coursework: CourseWork, where to search for submissions
    :param args: args from command line
    :return: list of submissions
    :raises: EntityNotFound if could not find any submissions with provided arguments
    """

    if args.submission_id is not None:
        return [classroom_service.get_submission_by_uuid(course_id=course.id, coursework_id=coursework.id,
                                                         submission_uuid=args.submission_id)]

    return classroom_service.get_submissions(course_id=course.id, coursework_id=coursework.id)


def parse_grading_schema(grading_schema_path: str) -> GradingSchema:
    """
    Parse grading schema from YAML file
    :param grading_schema_path: path to YAML file
    :return: GradingSchema instance
    """

    with open(grading_schema_path, 'r', encoding='utf8') as grading_schema_file:
        grading_schema_dict = yaml.safe_load(grading_schema_file)

    rules = []
    for rule in grading_schema_dict['schema']['rules'].values():
        rules.append(GradingRule(**rule))

    return GradingSchema(name=grading_schema_dict['schema']['name'],
                         total_grade=grading_schema_dict['schema']['total_grade'],
                         rules=rules)


def grade_submission(submission: Submission, grading_schema: GradingSchema, working_dir: str) -> GradingResult:
    """Grade submission by"""

    gained_grade = 0
    tester_output = ""

    for rule in grading_schema.rules:
        test_path = os.path.join(TMP_TESTS_DIR, f"{rule.task_name}_tests.py")

        script_destination = os.path.join(working_dir, rule.filename)

        if not os.path.exists(script_destination):
            tester_output += f"File {rule.filename} was not found\n"
            continue

        result = run_tests_for(script_path=script_destination, test_script_path=test_path)

        result.stream.seek(0)
        tester_output += result.stream.read()

        success_percentage = result.gained_points / result.total_points

        gained_grade += success_percentage * rule.weight * grading_schema.total_grade / 100

    return GradingResult(submission=submission, grade=gained_grade, comment=tester_output)


def export_grades_and_comments(results: list[GradingResult], students: list[Student], export_file_path: str) -> None:
    """
    Write all grades and comments to grading to CSV file
    :param results: list of GradingResult for each submission
    :param students: list of Students of course
    :param export_file_path: path to export file
    :return: None
    """

    with open(export_file_path, 'w', encoding='utf8') as export_file:
        export_file.write("Student,Grade,Comment\n")
        for result in results:
            export_file.write(f"{students[result.submission.user_id].name},{result.grade},\"{result.comment}\"\n")


TMP_SUBMISSIONS_DIR = '.submissions'
TMP_TESTS_DIR = '.rgrader_tests'


def collect_submissions(
    course: Course,
    coursework: CourseWork,
    submissions: list[Submission],
    students: list[Student],
    args: argparse.ArgumentParser,
) -> None:
    """Download all submissions attachments for this grading"""

    if not os.path.exists(TMP_SUBMISSIONS_DIR):
        os.mkdir(TMP_SUBMISSIONS_DIR)

    if args.submission_id is not None:
        student = students[submissions[0].user_id]
        target_dir = os.path.join(TMP_SUBMISSIONS_DIR, coursework.title, student.name)
        get_submission(
            submission=submissions[0],
            student=student,
            target_dir=target_dir,
            extract_if_single_directory=args.extract_if_single_dir,
        )
    else:
        get_course_work(
            course_info={"id": course.id},
            course_work=coursework,
            target_dir=TMP_TESTS_DIR,
            extract_if_single_directory=args.extract_if_single_dir,
        )


def collect_test_scripts(grading_schema: GradingSchema) -> dict[GradingRule, str]:
    """Copy tests, specified in grading_schema to TMP_TESTS_DIR
    :returns dict: of Grading rule and path to copied test script was copied
    """

    result = {}

    if not os.path.exists(TMP_TESTS_DIR):
        os.mkdir(TMP_TESTS_DIR)

    for rule in grading_schema.rules:
        test_ext = os.path.splitext(rule.test_path)[1]

        copied_tests_path = os.path.join(TMP_TESTS_DIR, f"{rule.task_name}_tests{test_ext}")
        shutil.copyfile(rule.test_path, copied_tests_path)

        result[rule] = copied_tests_path

    return result

def read_submission_from_dir(path: str) -> list[Submission]:
    """
    Read submissions from a directory
    :param path: path to base folder of coursework
    """

    if not os.path.exists(os.path.join(path, '.course_work.yaml')):
        raise ValueError(f"There is not .course_work.yaml in the given path {path}")

    submissions = []
    for submission_dir in os.listdir(path):
        base_path = os.path.join(path, submission_dir)
        if not os.path.isdir(base_path):
            continue

        if not os.path.exists(os.path.join(base_path, '.submission.yaml')):
            continue

        submissions.append(read_sumbission_yaml(base_path))

    return submissions


def grade_submissions(args: argparse.Namespace) -> None:
    """Grade submission from classroom"""

    course = get_current_course()
    students = classroom_service.get_students(course_id=course.id)
    grading_schema = parse_grading_schema(args.grading_schema)
    collect_test_scripts(grading_schema)

    if args.submission_folder is None:
        coursework = classroom_service.get_coursework_by_uuid(course_id=course.id, coursework_uuid=args.coursework)
        submissions = get_submissions_to_grade(course=course, coursework=coursework, args=args)
        collect_submissions(course, coursework, submissions, students, args)
    else:
        coursework = read_course_work_yaml(args.submission_folder)
        submissions = read_submission_from_dir(args.submission_folder)

    results = []

    for submission in submissions:

        if args.submission_folder is None:
            working_dir = os.path.join(TMP_SUBMISSIONS_DIR, coursework.title, students[submission.user_id].name)
        else:
            working_dir = os.path.join(coursework.title, students[submission.user_id].name)

        results.append(grade_submission(submission=submission, grading_schema=grading_schema, working_dir=working_dir))

    if args.export_file is not None:
        export_grades_and_comments(results, students=students, export_file_path=args.export_file)

    if args.submission_folder is None:
        shutil.rmtree(TMP_SUBMISSIONS_DIR)

    shutil.rmtree(TMP_TESTS_DIR)



"""Decorators for testing"""
from functools import wraps

from .utils import capture_io


def add_points(points: float):
    def decorator(test_function: callable):
        setattr(test_function, 'points_increment', points)
        return test_function

    return decorator


def run_solution(inputs: list):
    def decorator(test_function: callable):
        @wraps(test_function)
        def wrapper(class_self):
            from .globals import testing_script_path
            outputs = capture_io(testing_script_path, map(str, inputs))
            return test_function(class_self, outputs)

        return wrapper

    return decorator



"""Main grader module"""
import argparse

from .classroom.commands import run_no_args_argparse, get_course, choose_course_work, choose_course_work_argparse
from .classroom.commands import setup_grade_submissions, grade_submissions
from .classroom.interactive import run_tests_interactive, run_tests_interactive_argument_parser
from .test_runner import run_tests_setup_argparse, run_tests

ACTIONS = {
    'rt': (run_tests_setup_argparse, run_tests),
    'rti': (run_no_args_argparse, run_tests_interactive),
    'gc': (run_no_args_argparse, get_course),
    'submissions': (setup_grade_submissions, grade_submissions),
    'ga': (choose_course_work_argparse, choose_course_work)
}


def grade():
    """
    Run tests for solution script with specified test file and show grade
    """
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("action", help="Action to perform", choices=ACTIONS)

    for setup_function, _ in ACTIONS.values():
        setup_function(argument_parser)

    args = argument_parser.parse_args()
    action = args.action

    if action not in ACTIONS:
        print(f"There is no such action `{action}`. Available actions are: {', '.join(ACTIONS.keys())}")

    try:
        ACTIONS[action][1](args)
    except Exception as ex:
        print(str(ex))



"""Google Drive API client"""
import io
import shutil

import google_auth_oauthlib.flow
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.http import MediaIoBaseDownload


class GoogleDriveService(object):
    """Google Drive client"""

    SCOPES = [
        "https://www.googleapis.com/auth/drive.readonly"
    ]

    def __init__(self, credentials):
        self._service = build('drive', 'v3', credentials=credentials)

    def download_file(self, file_id: str, destination: str) -> None:
        """Download file from Google Drive by file_id"""

        request = self._service.files().get_media(fileId=file_id)

        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()

        fh.seek(0)
        with open(destination, 'wb') as f:
            shutil.copyfileobj(fh, f, length=131072)

    def _get_credentials(self):
        flow = InstalledAppFlow.from_client_secrets_file(self.client_secret, self.scopes)
        self.creds = flow.run_local_server(port=8080)
        return self.creds



"""Classroom API connector"""

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from .model import Course, CourseWork, Attachment, Submission, Student
from ..exceptions import EntityNotFound


class ClassroomConnector:
    """Connector to classroom API"""

    SCOPES = ["https://www.googleapis.com/auth/classroom.courses.readonly",
              "https://www.googleapis.com/auth/classroom.coursework.students",
              "https://www.googleapis.com/auth/classroom.rosters"]

    def __init__(self, credentials: str) -> None:
        """
        Initialize classroom connector

        :param credentials_file str: path to credentials file
        """

        self._service = build("classroom", "v1", credentials=credentials)
        self.courses = []

    def get_courses(self) -> list[Course]:
        """
        Return dict of (course_name, course_info)
        """

        try:
            page_token = None

            while True:
                response = self._service.courses().list(pageToken=page_token, pageSize=100).execute()

                page_courses = [Course(id=course['id'],
                                       name=course['name'],
                                       section=course.get('section'),
                                       description=course.get('descriptionHeading')) for course in
                                response.get("courses", [])]
                self.courses.extend(page_courses)
                page_token = response.get("nextPageToken", None)

                if not page_token:
                    break

            return self.courses
        except HttpError as error:
            print(f"An error occurred: {error}")
            raise error

    def get_courseworks(self, course_id: str):
        """
        Return list of all courseworks from specified course

        :param course_id: course from which list courseworkd
        :return:
        """

        coursework = []
        page_token = None

        try:
            while True:
                response = self._service.courses().courseWork().list(
                    pageToken=page_token, courseId=course_id,
                    pageSize=10
                ).execute()
                page_courseworks = [
                    CourseWork(id=coursework['id'],
                               title=coursework['title'],
                               description=coursework.get("description"),
                               course_id=course_id,
                               link=coursework['alternateLink'])
                    for coursework in response.get("courseWork", [])
                ]

                coursework.extend(page_courseworks)

                page_token = response.get("nextPageToken", None)
                if not page_token:
                    break
        except HttpError as error:
            print(f"An error occurred: {error}")

        return coursework

    def get_coursework_by_uuid(self, course_id, coursework_uuid):
        """
        Get coursework entity by course_id and coursework_uuid [id from the
        URL address in Classroom web client]
        :param course_id: id of course
        :param coursework_uuid: uuid of coursework [id from the
        URL address in Classroom web client]
        :return: coursework entity
        :raises: EntityNotFound if search failed
        """

        coursework = [coursework for coursework in self.get_courseworks(course_id=course_id)
                      if coursework.link.split('/')[6] == coursework_uuid]
        if not coursework:
            raise EntityNotFound(f"There is no coursework with id {coursework_uuid}")

        coursework = coursework[0]

        return coursework

    def get_submissions(self, course_id: str, coursework_id: str) -> list:
        """
        Return all submissions from course with [course_id] for assignment [coursework_id]
        :param course_id:
        :param coursework_id:
        :return:
        """

        submissions = []
        page_token = None

        try:
            while True:
                coursework = self._service.courses().courseWork()
                response = (
                    coursework.studentSubmissions()
                    .list(
                        pageToken=page_token,
                        courseId=course_id,
                        courseWorkId=coursework_id,
                        pageSize=10,
                    )
                    .execute()
                )

                for submission in response.get("studentSubmissions", []):
                    attachments = []

                    for attachment in submission.get('assignmentSubmission', {}).get('attachments', []):
                        attachment = attachment['driveFile']
                        attachments.append(Attachment(id=attachment['id'], title=attachment['title'],
                                                      link=attachment['alternateLink']))
                    submissions.append(Submission(id=submission['id'], user_id=submission['userId'],
                                                  attachments=attachments, course_id=course_id,
                                                  coursework_id=coursework_id, link=submission['alternateLink'],
                                                  creation_time=submission.get("creationTime"),
                                                  update_time=submission.get("updateTime")))

                page_token = response.get("nextPageToken", None)

                if not page_token:
                    break

        except HttpError as error:
            print(f"An error occurred: {error}")
            submissions = None

        return submissions

    def get_submission_by_uuid(self, course_id: str, coursework_id: str, submission_uuid: str) -> Submission:
        """
        Get submission entity by course_id, coursework_id and submission_uuid
        [id from the URL address in Classroom web client]
        :param course_id: id of course
        :param coursework_id: id of coursework
        :param submission_uuid: id of submission [id from the
        URL address in Classroom web client]
        :return: submission entity
        :raises: EntityNotFound if search failed
        """

        submission = [submission for submission in self.get_submissions(course_id=course_id,
                                                                        coursework_id=coursework_id)
                      if submission.link.split('/')[-1] == submission_uuid
                      ]
        if not submission:
            raise EntityNotFound(f"There is no coursework with id {submission_uuid}")

        submission = submission[0]

        return submission

    def grade_submission(self, submission: Submission, grade: int) -> None:
        """Grade submission with grade"""

        studentSubmission = {
            'assignedGrade': 99,
            'draftGrade': 80
        }
        self._service.courses().courseWork().studentSubmissions().patch(
            courseId=submission.course_id,
            courseWorkId=submission.coursework_id,
            id=submission.id,
            updateMask='assignedGrade,draftGrade',
            body=studentSubmission).execute()

    def get_students(self, course_id: str):
        """
        Return list of all students from specified course

        :param course_id: course from which list students
        :return:
        """

        students = {}
        page_token = None

        try:
            while True:

                response = self._service.courses().students().list(
                    pageToken=page_token, courseId=course_id,
                    pageSize=10
                ).execute()

                for student in response.get("students", []):
                    students[student['userId']] = Student(id=student["userId"],
                                                          name=student["profile"]["name"]["fullName"], )

                page_token = response.get("nextPageToken", None)
                if not page_token:
                    break
        except HttpError as error:
            print(f"An error occurred: {error}")

        return students



import unittest
from typing import Type
from unittest import TestCase

from .decorators import add_points, run_solution


class WrongStructure(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


def func_generator(name: str, inputs: list[str], exp_output: str, points: str, test_msg: str):
    try:
        @add_points(float(points))
        @run_solution(inputs)
        def testing_func(self, output: list[str]):
            self.assertEqual(output, exp_output, test_msg)

        testing_func.__name__ = name

        return testing_func
    except ValueError as e:
        print(e, "'points' must be a number with floating point")


def txt_to_func(test_file: str) -> dict:
    with open(test_file, "r", encoding="utf-8") as file:

        func_dict = {}
        file = file.read()

        # handle wrong test file structure
        input_lines = file.count("INPUT")
        output_lines = file.count("OUTPUT")
        cases_num = file.count("# test_")

        if input_lines != output_lines or output_lines != cases_num or input_lines != cases_num:
            raise WrongStructure(
                "Each test case should look like:\n# test_<name> <points>\n\
<optional test comment>\nINPUT\n<input per line>\nOUTPUT\n<output per line>")

        test_cases = filter(lambda x: bool(x), file.split("# "))

        for test in test_cases:
            test = test.strip().split("\n")

            if len(test) < 3:
                raise WrongStructure("Not enough elements in test case")

            if " " in test[0]:
                name, points = test[0].split(" ")
            else:
                name, points = test[0], 0

            if test[1] != "INPUT" and test[2] != "INPUT":
                raise WrongStructure(
                    f"INPUT goes right after test's name. It should look like:\n# {name} {points}\nINPUT")

            test_msg = ""
            if test[2] == "INPUT":
                test_msg = test[1]

            inputs_start = test.index("INPUT")

            if "OUTPUT" not in test:
                raise WrongStructure(f"No OUTPUT for {name}.")

            output_start = test.index("OUTPUT")

            if output_start - inputs_start == 1:
                raise WrongStructure(f"No inputs are given for {name}.")

            if len(test) - 1 == output_start:
                raise WrongStructure(f"No outputs are given for {name}.")

            inputs = test[inputs_start + 1:output_start]
            outputs = "\n".join(test[output_start + 1:])

            func_dict[name] = func_generator(name, inputs, outputs, points, test_msg)

        return func_dict


def create_unittest(test_file: str) -> Type[TestCase]:
    return type("tests", (unittest.TestCase,), txt_to_func(test_file))



from setuptools import setup, find_packages

setup(
    name="rgrader",
    version='1.5.6',
    packages=find_packages(),
    description="A simple grader for python problems",
    install_requires=[
        'inquirer==3.3.0',
        "google-api-core==2.19.0",
        "google-api-python-client==2.131.0",
        "google-auth==2.29.0",
        "google-auth-httplib2==0.2.0",
        "google-auth-oauthlib==1.2.0",
        "googleapis-common-protos==1.63.0",
        "PyYAML==6.0.1",
        "pycurses==0.0.25"
    ],
    entry_points={
        'console_scripts': [
            "grade = rgrader:grade"
        ]
    }
)



"""Models for interacting with Google Classroom API."""
from dataclasses import dataclass


@dataclass
class Course:
    """Course in classroom"""
    id: str
    name: str
    section: str = None
    description: str = None


@dataclass
class CourseWork:
    """CourseWork in classroom"""
    id: str
    title: str
    description: str
    course_id: str
    link: str


@dataclass
class Attachment:
    """Attachment to google classroom submission"""
    id: str
    title: str
    link: str


@dataclass
class Submission:
    """Submission in classroom"""
    id: str
    user_id: str
    course_id: str
    coursework_id: str
    creation_time: str
    update_time: str
    attachments: list[Attachment]
    link: str


@dataclass
class GradingRule:
    """
    Keeps data on testing of separate entry in submission
    (i.e. file in submitted archive )
    """

    filename: str
    weight: str
    test_path: str

    @property
    def task_name(self) -> str:
        return self.filename.split('.')[0]

    def __hash__(self) -> int:
        return hash(self.test_path)


@dataclass
class GradingSchema:
    """
    Contains data on how to test and grade submission
    Attributes:
        name: The name of grading scheme
        total_grade: The total grade of the grading

    """

    name: str
    total_grade: int
    rules: list[GradingRule]

    def __post_init__(self):
        """
        Validate passed data.
        1. total_grade > 0
        2. test rules weights sum up to 100
        """
        if self.total_grade <= 0:
            raise ValueError("Total grade must be greater than 0")
        if sum(rule.weight for rule in self.rules) != 100:
            raise ValueError("Weights in grading rules must sum up to 100")


@dataclass
class GradingResult:
    """Keeps data about grade and comment to submission"""
    submission: Submission
    grade: float
    comment: str


@dataclass
class Student:
    id: str
    name: str
